@@ -12,12 +12,18 @@
 #include "fs.h"
 #include "codepage.h"
 
+#ifdef TEST_SIZE
+#include "fsP.h"
+#endif
+
 static int recursive;
 static int wide;
 static int all;
 static int concise;
 static int fast=0;
-
+#if 0
+static int testmode = 0;
+#endif
 static char *dirPath;
 static char currentDrive;
 static Stream_t *currentDir;
@@ -151,7 +157,7 @@ static const char *dotted_num(mt_size_t num, int width, char **buf)
 		dstp[1] = srcp[1];
 		dstp[2] = srcp[2];
 		if (dstp + 3 < (*buf) + len)
-			/* use spaces instead of dots: they place both
+			/* use spaces instead of dots: they please both
 			 * Americans and Europeans */
 			dstp[3] = ' ';		
 		srcp += 3;
@@ -161,25 +167,30 @@ static const char *dotted_num(mt_size_t num, int width, char **buf)
 	return (*buf) + len-width;
 }
 
-static inline void print_volume_label(Stream_t *Dir, char drive)
+static inline int print_volume_label(Stream_t *Dir, char drive)
 {
 	Stream_t *Stream = GetFs(Dir);
 	direntry_t entry;
 	DeclareThis(FsPublic_t);
 	char shortname[13];
 	char longname[VBUFSIZE];
+	int r;
 
 	RootDir = OpenRoot(Stream);
 	if(concise)
-		return;
+		return 0;
 	
 	/* find the volume label */
 
-	initializeDirentry(&entry, Dir);
-	if(vfat_lookup(&entry, 0, 0, ACCEPT_LABEL | MATCH_ANY,
-		       shortname, longname) )
+	initializeDirentry(&entry, RootDir);
+	if((r=vfat_lookup(&entry, 0, 0, ACCEPT_LABEL | MATCH_ANY,
+			  shortname, longname)) ) {
+		if (r == -2) {
+			/* I/O Error */
+			return -1;
+		}
 		printf(" Volume in drive %c has no label", drive);
-	else if (*longname)
+	} else if (*longname)
 		printf(" Volume in drive %c is %s (abbr=%s)",
 		       drive, longname, shortname);
 	else
@@ -189,6 +200,7 @@ static inline void print_volume_label(Stream_t *Dir, char drive)
 		printf("\n Volume Serial Number is %04lX-%04lX",
 		       (This->serial_number >> 16) & 0xffff, 
 		       This->serial_number & 0xffff);
+	return 0;
 }
 
 
@@ -210,14 +222,14 @@ static void printSummary(int files, mt_size_t bytes)
 	}
 }
 
-static void leaveDirectory(void);
+static void leaveDirectory(int haveError);
 
-static void leaveDrive(void)
+static void leaveDrive(int haveError)
 {
 	if(!currentDrive)
 		return;
-	leaveDirectory();
-	if(!concise) {
+	leaveDirectory(haveError);
+	if(!concise && !haveError) {
 		char *s1;
 
 		if(dirsOnDrive > 1) {
@@ -228,6 +240,12 @@ static void leaveDrive(void)
 			mt_off_t bytes = getfree(RootDir);
 			printf("                  %s bytes free\n\n",
 			       dotted_num(bytes,17, &s1));
+#ifdef TEST_SIZE
+			((Fs_t*)GetFs(RootDir))->freeSpace = 0;
+			bytes = getfree(RootDir);
+			printf("                  %s bytes free\n\n",
+			       dotted_num(bytes,17, &s1));
+#endif
 		}
 		if(s1)
 			free(s1);
@@ -237,47 +255,58 @@ static void leaveDrive(void)
 }
 
 
-static void enterDrive(Stream_t *Dir, char drive)
+static int enterDrive(Stream_t *Dir, char drive)
 {
+	int r;
 	if(currentDrive == drive)
-		return; /* still the same */
+		return 0; /* still the same */
 	
-	leaveDrive();
-	
-	print_volume_label(Dir, drive);
-
+	leaveDrive(0);
 	currentDrive = drive;
+	
+	r = print_volume_label(Dir, drive);
+	if (r)
+		return r;
+
+
 	bytesOnDrive = 0;
 	filesOnDrive = 0;
 	dirsOnDrive = 0;
+	return 0;
 }
 
 static char *emptyString="<out-of-memory>";
 
-static void leaveDirectory(void)
+static void leaveDirectory(int haveError)
 {
 	if(!currentDir)
 		return;
-	if(dirPath && dirPath != emptyString)
-		free(dirPath);
-	if(wide)
-		putchar('\n');
 
-	if(!concise)
-		printSummary(filesInDir, bytesInDir);
+	if (!haveError) {
+		if(dirPath && dirPath != emptyString)
+			free(dirPath);
+		if(wide)
+			putchar('\n');
+		
+		if(!concise)
+			printSummary(filesInDir, bytesInDir);
+	}
 	FREE(&currentDir);
 }
 
-static void enterDirectory(Stream_t *Dir)
+static int enterDirectory(Stream_t *Dir)
 {
+	int r;
 	char drive;
 	if(currentDir == Dir)
-		return; /* still the same directory */
+		return 0; /* still the same directory */
 
-	leaveDirectory();
+	leaveDirectory(0);
 
 	drive = getDrive(Dir);
-	enterDrive(Dir, drive);
+	r=enterDrive(Dir, drive);
+	if(r)
+		return r;
 	currentDir = COPY(Dir);
 
 	dirPath = getPwd(getDirentry(Dir));
@@ -296,14 +325,15 @@ static void enterDirectory(Stream_t *Dir)
 	dirsOnDrive++;
 	bytesInDir = 0;
 	filesInDir = 0;
+	return 0;
 }
 
-
 static int list_file(direntry_t *entry, MainParam_t *mp)
 {
 	unsigned long size;
 	int i;
 	int Case;
+	int r;
 
 	if(!all && (entry->dir.attr & 0x6))
 		return 0;
@@ -311,7 +341,9 @@ static int list_file(direntry_t *entry, MainParam_t *mp)
 	if(concise && isSpecial(entry->name))
 		return 0;
 
-	enterDirectory(entry->Dir);
+	r=enterDirectory(entry->Dir);
+	if (r)
+		return ERROR_ONE;
 	if (wide) {
 		if(filesInDir % 5)
 			putchar(' ');				
@@ -385,6 +417,7 @@ static int list_file(direntry_t *entry, MainParam_t *mp)
 
 static int list_non_recurs_directory(direntry_t *entry, MainParam_t *mp)
 {
+	int r;
 	/* list top-level directory
 	 *   If this was matched by wildcard in the basename, list it as
 	 *   file, otherwise, list it as directory */
@@ -395,7 +428,9 @@ static int list_non_recurs_directory(direntry_t *entry, MainParam_t *mp)
 		/* no wildcard, list it as directory */
 		MainParam_t subMp;
 
-		enterDirectory(mp->File);
+		r=enterDirectory(mp->File);
+		if(r)
+			return ERROR_ONE;
 
 		subMp = *mp;
 		subMp.dirCallback = subMp.callback;
@@ -423,16 +458,31 @@ static int list_recurs_directory(direntry_t *entry, MainParam_t *mp)
 	return ret | mp->loop(mp->File, &subMp, "*");
 }
 
+#if 0
+static int test_directory(direntry_t *entry, MainParam_t *mp)
+{
+	Stream_t *File=mp->File;
+	Stream_t *Target;
+	char errmsg[80];
 
+	if ((Target = SimpleFileOpen(0, 0, "-",
+				     O_WRONLY,
+				     errmsg, 0, 0, 0))) {
+		copyfile(File, Target);
+		FREE(&Target);
+	}
+	return GOT_ONE;
+}
+#endif
 
 static void usage(void)
 {
 		fprintf(stderr, "Mtools version %s, dated %s\n",
 			mversion, mdate);
-		fprintf(stderr, "Usage: %s: [-V] [-w] [-a] msdosdirectory\n",
+		fprintf(stderr, "Usage: %s: [-V] [-w] [-a] [-b] [-s] [-f] msdosdirectory\n",
 			progname);
 		fprintf(stderr,
-			"       %s: [-V] [-w] [-a] msdosfile [msdosfiles...]\n",
+			"       %s: [-V] [-w] [-a] [-b] [-s] [-f] msdosfile [msdosfiles...]\n",
 			progname);
 		exit(1);
 }
@@ -450,7 +500,7 @@ void mdir(int argc, char **argv, int type)
 	recursive = 0;
 	wide = all = 0;
 					/* first argument */
-	while ((c = getopt(argc, argv, "waXfd/")) != EOF) {
+	while ((c = getopt(argc, argv, "waXbfds/")) != EOF) {
 		switch(c) {
 			case 'w':
 				wide = 1;
@@ -458,10 +508,12 @@ void mdir(int argc, char **argv, int type)
 			case 'a':
 				all = 1;
 				break;
+			case 'b':
 			case 'X':
 				concise = 1;
 				/*recursive = 1;*/
 				break;
+			case 's':
 			case '/':
 				recursive = 1;
 				break;
@@ -471,6 +523,11 @@ void mdir(int argc, char **argv, int type)
 			case 'd':
 				debug = 1;
 				break;
+#if 0
+			case 't': /* test mode */
+				testmode = 1;
+				break;
+#endif
 			default:
 				usage();
 		}
@@ -489,7 +546,13 @@ void mdir(int argc, char **argv, int type)
 	currentDir = 0;
 	RootDir = 0;
 	dirPath = 0;
-	if(recursive) {
+#if 0
+	if (testmode) {
+		mp.lookupflags = ACCEPT_DIR | NO_DOTS;
+		mp.dirCallback = test_directory;
+	} else 
+#endif
+		if(recursive) {
 		mp.lookupflags = ACCEPT_DIR | DO_OPEN_DIRS | NO_DOTS;
 		mp.dirCallback = list_recurs_directory;
 	} else {
@@ -500,7 +563,7 @@ void mdir(int argc, char **argv, int type)
 	mp.longname = longname;
 	mp.shortname = shortname;
 	ret=main_loop(&mp, argv + optind, argc - optind);
-	leaveDirectory();
-	leaveDrive();
+	leaveDirectory(ret);
+	leaveDrive(ret);
 	exit(ret);
 }
